@@ -24,6 +24,9 @@ continue  finally   is        return
 def       for       lambda    try
 """.split()
 
+
+# {{{ data model
+
 class Argument:
     def __init__(self, name, semantics, base_type, ptr):
         self.name = name
@@ -31,6 +34,7 @@ class Argument:
         self.base_type = base_type
         self.ptr = ptr
 
+
 class CallbackArgument:
     def __init__(self, name, return_base_type, return_ptr, args):
         self.name = name
@@ -38,6 +42,7 @@ class CallbackArgument:
         self.return_ptr = return_ptr
         self.args = args
 
+
 class Method:
     def __init__(self, cls, name, c_name,
             return_semantics, return_base_type, return_ptr,
@@ -73,35 +78,53 @@ class Method:
     def __repr__(self):
         return "<method %s>" % self.c_name
 
-
+# }}}
 
 
 PART_TO_CLASSES = {
-        "part1": [
-        "basic_set_list", "set_list", "aff_list", "pw_aff_list", "band_list",
-        "printer",  "mat", "vec", "id",
-        "aff", "pw_aff",
-        "multi_aff", "multi_pw_aff", "pw_multi_aff", "union_pw_multi_aff",
+        # If you change this, change:
+        # - src/wrapper/wrap_isl.hpp to add WRAP_CLASS(...)
+        # - src/wrapper/wrap_isl_partN.hpp to add MAKE_WRAP(...)
+        # - doc/reference.rst
 
-         "constraint", "space", "local_space",
-         ],
+        "part1": [
+            # lists
+            "id_list",
+            "basic_set_list", "set_list", "aff_list", "pw_aff_list", "band_list",
+            "ast_expr_list", "ast_node_list",
+
+            # maps
+            "id_to_ast_expr",
+
+            # others
+            "printer",  "val", "multi_val", "vec", "mat",
+            "aff", "pw_aff",
+            "multi_aff", "multi_pw_aff", "pw_multi_aff", "union_pw_multi_aff",
+
+            "id",
+            "constraint", "space", "local_space",
+        ],
 
         "part2": [
-        "basic_set", "basic_map",
-        "set", "map",
-        "union_map", "union_set",
-        "point", "vertex", "cell", "vertices",
+            "basic_set", "basic_map",
+            "set", "map",
+            "union_map", "union_set",
+            "point", "vertex", "cell", "vertices",
         ],
 
         "part3": [
-        "qpolynomial_fold", "pw_qpolynomial_fold",
-        "union_pw_qpolynomial_fold",
-        "union_pw_qpolynomial", "term",
-        "qpolynomial", "pw_qpolynomial",
+            "qpolynomial_fold", "pw_qpolynomial_fold",
+            "union_pw_qpolynomial_fold",
+            "union_pw_qpolynomial",
+            "qpolynomial", "pw_qpolynomial",
+            "term",
 
-        "band", "schedule",
+            "band", "schedule", "schedule_constraints",
 
-        "access_info", "flow", "restriction",
+            "access_info", "flow", "restriction",
+
+            "ast_expr", "ast_node", "ast_print_options",
+            "ast_build",
         ]
         }
 CLASSES = []
@@ -114,11 +137,16 @@ CLASS_MAP = {
         "options": "ctx",
         }
 
-ENUMS = ["isl_dim_type", "isl_fold"]
+ENUMS = ["isl_dim_type", "isl_fold",
+        "isl_ast_op_type", "isl_ast_expr_type",
+        "isl_ast_node_type"]
 
-SAFE_TYPES = ENUMS + ["int", "unsigned", "uint32_t", "size_t"]
+SAFE_TYPES = ENUMS + ["int", "unsigned", "uint32_t", "size_t", "double",
+        "long", "unsigned long"]
 SAFE_IN_TYPES = SAFE_TYPES + ["const char *", "char *"]
 
+# {{{ parser
+
 DECL_RE = re.compile(r"""
     ((?:\w+\s+)*) (\**) \s* (?# return type)
     (\w+) (?# func name)
@@ -139,6 +167,7 @@ STRUCT_DECL_RE = re.compile(r"(__isl_export\s+)?struct\s+([a-z_A-Z0-9]+)\s*;")
 ARG_RE = re.compile(r"^((?:\w+)\s+)+(\**)\s*(\w+)$")
 INLINE_SEMICOLON_RE = re.compile(r"\;[ \t]*(?=\w)")
 
+
 def filter_semantics(words):
     semantics = []
     other_words = []
@@ -155,8 +184,6 @@ def filter_semantics(words):
         return None, other_words
 
 
-
-
 def split_at_unparenthesized_commas(s):
     paren_level = 0
     i = 0
@@ -177,7 +204,6 @@ def split_at_unparenthesized_commas(s):
     yield s[last_start:i]
 
 
-
 def to_py_class(cls):
     if cls.startswith("isl_"):
         cls = cls[4:]
@@ -203,22 +229,22 @@ def to_py_class(cls):
     return result
 
 
-
-
 class Retry(RuntimeError):
     pass
 
+
 class BadArg(ValueError):
     pass
 
+
 class Undocumented(ValueError):
     pass
 
+
 class SignatureNotSupported(ValueError):
     pass
 
 
-
 def parse_arg(arg):
     if "(*" in arg:
         arg_match = FUNC_PTR_RE.match(arg)
@@ -249,10 +275,6 @@ def parse_arg(arg):
             ptr=arg_match.group(2).strip())
 
 
-
-
-
-
 class FunctionData:
     def __init__(self, include_dirs):
         self.classes_to_methods = {}
@@ -288,7 +310,6 @@ class FunctionData:
 
         i = 0
 
-        current_decl = ""
         while i < len(lines):
             l = lines[i].strip()
 
@@ -308,7 +329,7 @@ class FunctionData:
                     i += 1
 
                     l = lines[i].strip()
-            elif l.endswith("{") :
+            elif l.endswith("{"):
                 while True:
                     if "}" in l:
                         i += 1
@@ -353,7 +374,8 @@ class FunctionData:
         args = [i.strip()
                 for i in split_at_unparenthesized_commas(decl_match.group(4))]
 
-        if c_name == "ISL_ARG_DECL":
+        if c_name in ["ISL_ARG_DECL", "ISL_DECLARE_MULTI",
+                "ISL_DECLARE_LIST", "isl_ast_op_type_print_macro"]:
             return
 
         assert c_name.startswith("isl_"), c_name
@@ -368,6 +390,16 @@ class FunctionData:
         if found_class:
             name = name[len(cls)+1:]
 
+        # Don't be tempted to chop off "_val"--the "_val" versions of
+        # some methods are incompatible with the isl_int ones.
+        #
+        # (For example, isl_aff_get_constant() returns just the constant,
+        # but isl_aff_get_constant_val() returns the constant divided by
+        # the denominator.)
+        #
+        # To avoid breaking user code in non-obvious ways, the new
+        # names are carried over to the Python level.
+
         if not found_class:
             for fake_cls, cls in CLASS_MAP.items():
                 if name.startswith(fake_cls):
@@ -412,11 +444,10 @@ class FunctionData:
                 return_semantics, return_base_type, return_ptr,
                 args, is_exported=is_exported, is_constructor=is_constructor))
 
-
+# }}}
 
 
 def get_callback(cb_name, cb):
-    input_args = []
     body = []
     passed_args = []
 
@@ -424,7 +455,9 @@ def get_callback(cb_name, cb):
 
     for arg in cb.args[:-1]:
         if arg.base_type.startswith("isl_"):
-            assert arg.ptr == "*"
+            if arg.ptr != "*":
+                raise SignatureNotSupported("unsupported callback arg: %s %s" % (
+                    arg.base_type, arg.ptr))
             arg_cls = arg.base_type[4:]
 
             if arg.semantics is not SEM_TAKE:
@@ -435,7 +468,8 @@ def get_callback(cb_name, cb):
             body.append("""
                 std::auto_ptr<%(arg_cls)s> wrapped_arg_%(name)s(
                     new %(arg_cls)s(c_arg_%(name)s));
-                py::object arg_%(name)s(handle_from_new_ptr(wrapped_arg_%(name)s.get()));
+                py::object arg_%(name)s(
+                    handle_from_new_ptr(wrapped_arg_%(name)s.get()));
                 wrapped_arg_%(name)s.release();
                 """ % dict(
                     arg_cls=arg_cls,
@@ -484,7 +518,7 @@ def get_callback(cb_name, cb):
                 passed_args=", ".join(passed_args))
 
 
-
+# {{{ wrapper generator
 
 def write_wrapper(outf, meth):
     body = []
@@ -522,7 +556,10 @@ def write_wrapper(outf, meth):
             preamble.append(get_callback(cb_name, arg))
 
             docs.append(":param %s: callback(%s)"
-                    % (arg.name, ", ".join(sub_arg.name for sub_arg in arg.args if sub_arg.name != "user")))
+                    % (arg.name, ", ".join(
+                        sub_arg.name
+                        for sub_arg in arg.args
+                        if sub_arg.name != "user")))
 
         elif arg.base_type in SAFE_IN_TYPES and not arg.ptr:
             passed_args.append("arg_"+arg.name)
@@ -553,41 +590,84 @@ def write_wrapper(outf, meth):
             else:
                 raise SignatureNotSupported("int *")
 
-        elif arg.base_type == "isl_int" and arg.ptr == "*":
-            # assume it's meant as a return value
-            body.append("""
-                py::object arg_%(name)s(py::handle<>((PyObject *) Pympz_new()));
-                managed_int arg_mi_%(name)s;
+        elif arg.base_type == "isl_val" and arg.ptr == "*" and arg_idx > 0:
+            # {{{ val input argument
 
-                """ % dict(name=arg.name))
-            passed_args.append("&arg_mi_%s.m_data" % arg.name)
-            post_call.append("""
-                isl_int_get_gmp(arg_mi_%(name)s.m_data, Pympz_AS_MPZ(arg_%(name)s.ptr()));
-                """ % dict(name=arg.name))
-
-            extra_ret_vals.append("arg_%s" % arg.name)
-            extra_ret_descrs.append("%s (integer)" % arg.name)
-
-            arg_names.pop()
-
-        elif arg.base_type == "isl_int" and not arg.ptr:
-            input_args.append("py::object %s" % ("arg_"+arg.name))
+            arg_descr = ":param %s: :class:`Val`" % arg.name
+            input_args.append("py::object py_%s" % arg.name)
             checks.append("""
-                managed_int arg_mi_%(name)s;
-                {
-                  PyObject *converted;
-                  if (Pympz_convert_arg(arg_%(name)s.ptr(), &converted) == 0)
-                    throw py::error_already_set();
-                  py::handle<> converted_arg_%(name)s = py::handle<>(converted);
-                  isl_int_set_gmp(arg_mi_%(name)s.m_data,
-                    Pympz_AS_MPZ(converted_arg_%(name)s.get()));
-                }
-                """ % dict(name=arg.name, meth="%s_%s" % (meth.cls, meth.name)))
-            passed_args.append("arg_mi_%s.m_data" % arg.name)
+                std::auto_ptr<val> auto_arg_%(name)s;
+                py::extract<val *> ex_%(name)s(py_%(name)s);
+                isl_ctx *ctx_for_%(name)s =
+                    %(first_arg_base_type)s_get_ctx(arg_%(first_arg)s.m_data);
 
-            docs.append(":param %s: integer" % arg.name)
+                if (ex_%(name)s.check())
+                {
+                  val *arg_%(name)s = ex_%(name)s();
+                  if (!auto_arg_%(name)s->is_valid())
+                    throw isl::error(
+                      "passed invalid val for %(name)s");
+
+                  isl_val *tmp_ptr = isl_val_copy(arg_%(name)s->m_data);
+                  if (!tmp_ptr)
+                      throw isl::error("failed to copy arg %(name)s");
+                  auto_arg_%(name)s = std::auto_ptr<val>(new val(tmp_ptr));
+                }
+                else if (PyLong_Check(py_%(name)s.ptr()))
+                {
+                  long value = PyLong_AsLong(py_%(name)s.ptr());
+                  if (PyErr_Occurred())
+                    throw py::error_already_set();
+
+                  isl_val *tmp_ptr = isl_val_int_from_si(ctx_for_%(name)s, value);
+                  if (!tmp_ptr)
+                      throw isl::error("failed to create arg %(name)s from integer");
+                  auto_arg_%(name)s = std::auto_ptr<val>(new val(tmp_ptr));
+                }
+                """ % dict(
+                    name=arg.name,
+                    first_arg_base_type=meth.args[0].base_type,
+                    first_arg=meth.args[0].name,
+                    ))
+
+            if sys.version_info < (3,):
+                checks.append("""
+                    else if (PyInt_Check(py_%(name)s.ptr()))
+                    {
+                      isl_val *tmp_ptr = isl_val_int_from_si(ctx_for_%(name)s,
+                          PyInt_AsLong(py_%(name)s.ptr()));
+                      if (!tmp_ptr)
+                          throw isl::error("failed to create arg "
+                              "%(name)s from integer");
+                      auto_arg_%(name)s = std::auto_ptr<val>(new val(tmp_ptr));
+                    }
+                    """ % dict(
+                        name=arg.name,
+                        ))
+
+            checks.append("""
+                else
+                {
+                  throw isl::error("unrecognized argument for %(name)s");
+                }
+                """ % dict(
+                    name=arg.name,
+                    ))
+
+            if arg.semantics is None and arg.base_type != "isl_ctx":
+                raise Undocumented(meth)
+
+            if arg.semantics is SEM_TAKE:
+                post_call.append("auto_arg_%s.release();" % arg.name)
+
+            passed_args.append("auto_arg_%s->m_data" % arg.name)
+            docs.append(arg_descr)
+
+            # }}}
 
         elif arg.base_type.startswith("isl_") and arg.ptr == "*":
+            # {{{ isl types input arguments
+
             need_nonconst = False
 
             arg_cls = arg.base_type[4:]
@@ -598,7 +678,8 @@ def write_wrapper(outf, meth):
 
             checks.append("""
                 if (!arg_%(name)s.is_valid())
-                  throw isl::error("passed invalid arg to isl_%(meth)s for %(name)s");
+                  throw isl::error(
+                    "passed invalid arg to isl_%(meth)s for %(name)s");
                 """ % dict(name=arg.name, meth="%s_%s" % (meth.cls, meth.name)))
 
             copyable = arg_cls not in NON_COPYABLE
@@ -606,13 +687,17 @@ def write_wrapper(outf, meth):
                 if copyable:
                     checks.append("""
                         if (!arg_%(name)s.is_valid())
-                          throw isl::error("passed invalid arg to isl_%(meth)s for %(name)s");
+                          throw isl::error(
+                            "passed invalid arg to isl_%(meth)s for %(name)s");
                         std::auto_ptr<%(cls)s> auto_arg_%(name)s;
                         {
-                            isl_%(cls)s *tmp_ptr = isl_%(cls)s_copy(arg_%(name)s.m_data);
+                            isl_%(cls)s *tmp_ptr =
+                                isl_%(cls)s_copy(arg_%(name)s.m_data);
                             if (!tmp_ptr)
-                                throw isl::error("failed to copy arg %(name)s on entry to %(meth)s");
-                            auto_arg_%(name)s = std::auto_ptr<%(cls)s>(new %(cls)s(tmp_ptr));
+                                throw isl::error("failed to copy arg "
+                                    "%(name)s on entry to %(meth)s");
+                            auto_arg_%(name)s = std::auto_ptr<%(cls)s>(
+                                new %(cls)s(tmp_ptr));
                         }
                         """ % dict(
                             name=arg.name,
@@ -644,7 +729,11 @@ def write_wrapper(outf, meth):
 
             docs.append(arg_descr)
 
+            # }}}
+
         elif arg.base_type.startswith("isl_") and arg.ptr == "**":
+            # {{{ isl types output arguments
+
             if arg.semantics is not SEM_GIVE:
                 raise SignatureNotSupported("non-give secondary ptr return value")
 
@@ -658,19 +747,41 @@ def write_wrapper(outf, meth):
                 py::object py_ret_%(name)s;
                 if (ret_%(name)s)
                 {
-                  std::auto_ptr<%(ret_cls)s> auto_ret_%(name)s(new %(ret_cls)s(ret_%(name)s));
-                  py_ret_%(name)s = py::object(handle_from_new_ptr(auto_ret_%(name)s.get()));
+                  std::auto_ptr<%(ret_cls)s> auto_ret_%(name)s(
+                    new %(ret_cls)s(ret_%(name)s));
+                  py_ret_%(name)s = py::object(
+                    handle_from_new_ptr(auto_ret_%(name)s.get()));
                   auto_ret_%(name)s.release();
                 }
                 """ % dict(name=arg.name, ret_cls=ret_cls))
 
             extra_ret_vals.append("py_ret_%s" % arg.name)
-            extra_ret_descrs.append("%s (:class:`%s`)" % (arg.name, to_py_class(ret_cls)))
+            extra_ret_descrs.append(
+                    "%s (:class:`%s`)" % (arg.name, to_py_class(ret_cls)))
+
+            # }}}
 
         elif arg.base_type == "FILE" and arg.ptr == "*":
+            if sys.version_info >= (3,):
+                raise SignatureNotSupported(
+                        "arg type %s %s" % (arg.base_type, arg.ptr))
+
             passed_args.append("PyFile_AsFile(arg_%s.ptr())" % arg.name)
             input_args.append("py::object %s" % ("arg_"+arg.name))
-            docs.append(":param %s: :class:`file`-like" % arg.name)
+            docs.append(":param %s: :class:`file`-like "
+                    "(NOTE: This will cease to be supported in Python 3.)"
+                    % arg.name)
+
+        elif (arg.base_type == "void"
+                and arg.ptr == "*"
+                and arg.name == "user"):
+            body.append("Py_INCREF(arg_%s.ptr());" % arg.name)
+            passed_args.append("arg_%s.ptr()" % arg.name)
+            input_args.append("py::object %s" % ("arg_"+arg.name))
+            post_call.append("""
+                isl_%s_set_free_user(result, my_decref);
+                """ % meth.cls)
+            docs.append(":param %s: a user-specified Python object" % arg.name)
 
         else:
             raise SignatureNotSupported("arg type %s %s" % (arg.base_type, arg.ptr))
@@ -691,12 +802,14 @@ def write_wrapper(outf, meth):
 
     body += post_call
 
+    # {{{ return value processing
+
     if meth.return_base_type == "int" and not meth.return_ptr:
         body.append("""
             if (result == -1)
             {
               throw isl::error("call to isl_%(cls)s_%(name)s failed");
-            }""" % { "cls": meth.cls, "name": meth.name })
+            }""" % {"cls": meth.cls, "name": meth.name})
 
         if meth.name.startswith("is_") or meth.name.startswith("has_"):
             processed_return_type = "bool"
@@ -739,7 +852,8 @@ def write_wrapper(outf, meth):
             ret_descr = ":class:`%s` (self)" % to_py_class(ret_cls)
         else:
             processed_return_type = "py::object"
-            isl_obj_ret_val = "py::object(handle_from_new_ptr(new %s(result)))" % ret_cls
+            isl_obj_ret_val = \
+                    "py::object(handle_from_new_ptr(new %s(result)))" % ret_cls
 
             if extra_ret_vals:
                 isl_obj_ret_val = "py::make_tuple(%s, %s)" % (
@@ -793,6 +907,16 @@ def write_wrapper(outf, meth):
 
         ret_descr = "string"
 
+    elif (meth.return_base_type == "void"
+            and meth.return_ptr == "*"
+            and meth.name == "get_user"):
+
+        body.append("""
+            return py::object(py::handle<>(py::borrowed((PyObject *) result)));
+            """)
+        ret_descr = "a user-specified python object"
+        processed_return_type = "py::object"
+
     elif meth.return_base_type == "void" and not meth.return_ptr:
         if extra_ret_vals:
             processed_return_type = "py::object"
@@ -810,6 +934,8 @@ def write_wrapper(outf, meth):
         raise SignatureNotSupported("ret type: %s %s in %s" % (
             meth.return_base_type, meth.return_ptr, meth))
 
+    # }}}
+
     outf.write("""
         %s
         %s %s_%s(%s)
@@ -828,8 +954,10 @@ def write_wrapper(outf, meth):
 
     return arg_names, "\n".join(docs)
 
+# }}}
 
 
+# {{{ exposer generator
 
 def write_exposer(outf, meth, arg_names, doc_str, static_decls):
     func_name = "isl::%s_%s" % (meth.cls, meth.name)
@@ -864,7 +992,7 @@ def write_exposer(outf, meth, arg_names, doc_str, static_decls):
             static_decls.append("wrap_%s.staticmethod(\"%s\");\n" % (
                 wrap_class, exp_py_name))
 
-
+# }}}
 
 
 def write_wrappers(expf, wrapf, methods):
@@ -872,11 +1000,23 @@ def write_wrappers(expf, wrapf, methods):
     static_decls = []
 
     for meth in methods:
-        if meth.name.endswith("_si") or meth.name.endswith("_ui") and len([
-            meth2.name == meth.name[:-3]
-            for meth2 in methods]):
-            # no need to expose C integer versions of things
-            continue
+        #print "TRY_WRAP:", meth
+        if meth.name.endswith("_si") or meth.name.endswith("_ui"):
+            val_versions = [
+                    meth2
+                    for meth2 in methods
+                    if meth2.cls == meth.cls
+                    and (
+                        meth2.name == meth.name[:-3]
+                        or meth2.name == meth.name[:-3] + "_val"
+                        )
+                    ]
+
+            if val_versions:
+                # no need to expose C integer versions of things
+                print("SKIP (val version available): %s -> %s"
+                        % (meth, ", ".join(str(s) for s in val_versions)))
+                continue
 
         try:
             arg_names, doc_str = write_wrapper(wrapf, meth)
@@ -899,10 +1039,10 @@ def write_wrappers(expf, wrapf, methods):
     print("SKIP (%d undocumented methods): %s" % (len(undoc), ", ".join(undoc)))
 
 
-
 def gen_wrapper(include_dirs):
     fdata = FunctionData(["."] + include_dirs)
     fdata.read_header("isl_list.h")
+    fdata.read_header("isl/id.h")
     fdata.read_header("isl/space.h")
     fdata.read_header("isl/set.h")
     fdata.read_header("isl/map.h")
@@ -915,13 +1055,15 @@ def gen_wrapper(include_dirs):
     fdata.read_header("isl/vertices.h")
     fdata.read_header("isl/point.h")
     fdata.read_header("isl/constraint.h")
+    fdata.read_header("isl/val.h")
     fdata.read_header("isl/vec.h")
     fdata.read_header("isl/mat.h")
     fdata.read_header("isl/band.h")
     fdata.read_header("isl/schedule.h")
     fdata.read_header("isl/flow.h")
     fdata.read_header("isl/options.h")
-
+    fdata.read_header("isl/ast.h")
+    fdata.read_header("isl/ast_build.h")
 
     for part, classes in PART_TO_CLASSES.items():
         expf = open("src/wrapper/gen-expose-%s.inc" % part, "wt")
@@ -936,9 +1078,8 @@ def gen_wrapper(include_dirs):
         wrapf.close()
 
 
-
-
-
 if __name__ == "__main__":
     from os.path import expanduser
     gen_wrapper([expanduser("isl/include")])
+
+# vim: foldmethod=marker
