@@ -1,5 +1,8 @@
-from islpy._isl import *
-from islpy.version import *
+from __future__ import division
+
+
+from islpy._isl import *  # noqa
+from islpy.version import *  # noqa
 
 
 _CHECK_DIM_TYPES = [
@@ -11,10 +14,11 @@ EXPR_CLASSES = tuple(cls for cls in ALL_CLASSES
 
 _DEFAULT_CONTEXT = Context()
 
-def _add_functionality():
-    import islpy._isl as _isl
 
-    # {{{ generic initialization
+def _add_functionality():
+    import islpy._isl as _isl  # noqa
+
+    # {{{ generic initialization, pickling
 
     def obj_new_from_string(cls, s, context=None):
         """Construct a new object from :class:`str` s.
@@ -34,10 +38,16 @@ def _add_functionality():
         assert self._made_from_string
         del self._made_from_string
 
+    def generic_getinitargs(self):
+        prn = Printer.to_str(self.get_ctx())
+        getattr(prn, "print_"+self._base_name)(self)
+        return (prn.get_str(),)
+
     for cls in ALL_CLASSES:
         if hasattr(cls, "read_from_str"):
             cls.__new__ = staticmethod(obj_new_from_string)
             cls.__init__ = obj_bogus_init
+            cls.__getinitargs__ = generic_getinitargs
 
     # }}}
 
@@ -74,10 +84,17 @@ def _add_functionality():
             cls.__eq__ = obj_eq
             cls.__ne__ = obj_ne
 
-    def obj_lt(self, other): return self.is_strict_subset(other)
-    def obj_le(self, other): return self.is_subset(other)
-    def obj_gt(self, other): return other.is_strict_subset(self)
-    def obj_ge(self, other): return other.is_subset(self)
+    def obj_lt(self, other):
+        return self.is_strict_subset(other)
+
+    def obj_le(self, other):
+        return self.is_subset(other)
+
+    def obj_gt(self, other):
+        return other.is_strict_subset(self)
+
+    def obj_ge(self, other):
+        return other.is_subset(self)
 
     for cls in [BasicSet, BasicMap, Set, Map]:
         cls.__lt__ = obj_lt
@@ -118,8 +135,36 @@ def _add_functionality():
 
     # {{{ Space
 
+    def space_get_id_dict(self, dimtype=None):
+        """Return a dictionary mapping variable :class:`Id` instances to tuples
+        of (:class:`dim_type`, index).
+
+        :param dimtype: None to get all variables, otherwise
+            one of :class:`dim_type`.
+        """
+        result = {}
+
+        def set_dim_name(name, tp, idx):
+            if name in result:
+                raise RuntimeError("non-unique var id '%s' encountered" % name)
+            result[name] = tp, idx
+
+        if dimtype is None:
+            types = _CHECK_DIM_TYPES
+        else:
+            types = [dimtype]
+
+        for tp in types:
+            for i in range(self.dim(tp)):
+                name = self.get_dim_id(tp, i)
+                if name is not None:
+                    set_dim_id(name, tp, i)
+
+        return result
+
     def space_get_var_dict(self, dimtype=None):
-        """Return a dictionary mapping variable names to tuples of (:class:`dim_type`, index).
+        """Return a dictionary mapping variable names to tuples of
+        (:class:`dim_type`, index).
 
         :param dimtype: None to get all variables, otherwise
             one of :class:`dim_type`.
@@ -186,6 +231,7 @@ def _add_functionality():
 
     Space.create_from_names = staticmethod(space_create_from_names)
     Space.get_var_dict = space_get_var_dict
+    Space.get_id_dict = space_get_id_dict
 
     # }}}
 
@@ -200,7 +246,7 @@ def _add_functionality():
             New for :class:`Aff`
         """
         for i, coeff in enumerate(args):
-            self = self.set_coefficient(dim_tp, i, coeff)
+            self = self.set_coefficient_val(dim_tp, i, coeff)
 
         return self
 
@@ -215,7 +261,7 @@ def _add_functionality():
             New for :class:`Aff`
         """
         try:
-            iterable = iterable.iteritems()
+            iterable = iterable.items()
         except AttributeError:
             pass
 
@@ -224,10 +270,10 @@ def _add_functionality():
 
         for name, coeff in iterable:
             if name == 1:
-                self = self.set_constant(coeff)
+                self = self.set_constant_val(coeff)
             else:
                 tp, idx = name_to_dim[name]
-                self = self.set_coefficient(tp, idx, coeff)
+                self = self.set_coefficient_val(tp, idx, coeff)
 
         return self
 
@@ -270,6 +316,27 @@ def _add_functionality():
 
     # }}}
 
+    # {{{ Id
+
+    def id_new(cls, name, user=None, context=None):
+        if context is None:
+            context = _DEFAULT_CONTEXT
+
+        result = cls.alloc(context, name, user)
+        result._made_from_python = True
+        return result
+
+    def id_bogus_init(self, name, user=None, context=None):
+        assert self._made_from_python
+        del self._made_from_python
+
+    Id.__new__ = staticmethod(id_new)
+    Id.__init__ = id_bogus_init
+    Id.user = property(Id.get_user)
+    Id.name = property(Id.get_name)
+
+    # }}}
+
     # {{{ Constraint
 
     def eq_from_names(space, coefficients={}):
@@ -349,27 +416,42 @@ def _add_functionality():
 
     # {{{ common functionality
 
+    def obj_get_id_dict(self, dimtype=None):
+        """Return a dictionary mapping :class:`Id` instances to tuples of
+        (:class:`dim_type`, index).
+
+        :param dimtype: None to get all variables, otherwise
+            one of :class:`dim_type`.
+        """
+        return self.get_space().get_id_dict(dimtype)
+
     def obj_get_var_dict(self, dimtype=None):
-        """Return a dictionary mapping variable names to tuples of (:class:`dim_type`, index).
+        """Return a dictionary mapping variable names to tuples of
+        (:class:`dim_type`, index).
 
         :param dimtype: None to get all variables, otherwise
             one of :class:`dim_type`.
         """
         return self.get_space().get_var_dict(dimtype)
 
+    def obj_get_var_ids(self, dimtype):
+        """Return a list of :class:`Id` instances for :class:`dim_type` *dimtype*."""
+        return [self.get_dim_name(dimtype, i) for i in xrange(self.dim(dimtype))]
+
     def obj_get_var_names(self, dimtype):
         """Return a list of dim names (in order) for :class:`dim_type` *dimtype*."""
         return [self.get_dim_name(dimtype, i) for i in xrange(self.dim(dimtype))]
 
     for cls in ALL_CLASSES:
         if hasattr(cls, "get_space") and cls is not Space:
+            cls.get_id_dict = obj_get_id_dict
             cls.get_var_dict = obj_get_var_dict
+            cls.get_var_ids = obj_get_var_ids
             cls.get_var_names = obj_get_var_names
             cls.space = property(cls.get_space)
 
     # }}}
 
-
     # {{{ PwAff
 
     def pwaff_get_pieces(self):
@@ -378,6 +460,7 @@ def _add_functionality():
         """
 
         result = []
+
         def append_tuple(*args):
             result.append(args)
 
@@ -404,7 +487,7 @@ def _add_functionality():
 
     def _number_to_aff(template, num):
         number_aff = Aff.zero_on_domain(template.get_domain_space())
-        number_aff = number_aff.set_constant(num)
+        number_aff = number_aff.set_constant_val(num)
 
         if isinstance(template, PwAff):
             result = PwAff.empty(template.get_space())
@@ -449,7 +532,7 @@ def _add_functionality():
             return NotImplemented
 
     def aff_flordiv(self, other):
-        return self.scale_down(other).floor()
+        return self.scale_down_val(other).floor()
 
     for aff_class in [Aff, PwAff]:
         aff_class.__add__ = aff_add
@@ -459,11 +542,71 @@ def _add_functionality():
         aff_class.__mul__ = aff_mul
         aff_class.__rmul__ = aff_mul
         aff_class.__neg__ = aff_class.neg
-        aff_class.__mod__ = aff_class.mod
+        aff_class.__mod__ = aff_class.mod_val
         aff_class.__floordiv__ = aff_flordiv
 
     # }}}
 
+    # {{{ Val
+
+    def val_new(cls, src, context=None):
+        if context is None:
+            context = _DEFAULT_CONTEXT
+
+        if isinstance(src, (str, unicode)):
+            result = cls.read_from_str(context, src)
+        elif isinstance(src, (int, long)):
+            result = cls.int_from_si(context, src)
+        else:
+            raise TypeError("'src' must be int or string")
+
+        result._made_from_python = True
+        return result
+
+    def val_bogus_init(self, src, context=None):
+        assert self._made_from_python
+        del self._made_from_python
+
+    def val_rsub(self, other):
+        return -self + other
+
+    def val_nonzero(self):
+        return not self.is_zero()
+
+    def val_repr(self):
+        return "%s(\"%s\")" % (type(self).__name__, self.to_str())
+
+    def val_to_python(self):
+        if not self.is_int():
+            raise ValueError("can only convert integer Val to python")
+
+        return long(self.to_str())
+
+    Val.__new__ = staticmethod(val_new)
+    Val.__init__ = val_bogus_init
+    Val.__add__ = Val.add
+    Val.__radd__ = Val.add
+    Val.__sub__ = Val.sub
+    Val.__rsub__ = val_rsub
+    Val.__mul__ = Val.mul
+    Val.__rmul__ = Val.mul
+    Val.__neg__ = Val.neg
+    Val.__mod__ = Val.mod
+    Val.__bool__ = Val.__nonzero__ = val_nonzero
+
+    Val.__lt__ = Val.lt
+    Val.__gt__ = Val.gt
+    Val.__le__ = Val.le
+    Val.__ge__ = Val.ge
+    Val.__eq__ = Val.eq
+    Val.__ne__ = Val.ne
+
+    Val.__repr__ = val_repr
+    Val.__str__ = Val.to_str
+    Val.to_python = val_to_python
+
+    # }}}
+
     # {{{ add automatic 'self' upcasts
 
     # note: automatic upcasts for method arguments are provided through
@@ -476,8 +619,8 @@ def _add_functionality():
 
     def add_upcasts(basic_class, special_class, upcast_method):
         from functools import update_wrapper
-
         from inspect import ismethod
+
         for method_name in dir(special_class):
             # do not overwrite existing methods
             if hasattr(basic_class, method_name):
@@ -485,7 +628,15 @@ def _add_functionality():
 
             method = getattr(special_class, method_name)
 
-            if ismethod(method):
+            my_ismethod = ismethod(method)
+            for meth_superclass in type(method).__mro__:
+                if "function" in meth_superclass.__name__:
+                    # inspect.ismethod does not work on Boost.Py callables in Py3,
+                    # hence this hack.
+                    my_ismethod = True
+                    break
+
+            if my_ismethod:
                 def make_wrapper(method, upcast):
                     # This function provides a scope in which method and upcast
                     # are not changed from one iteration of the enclosing for
@@ -607,13 +758,9 @@ def _add_functionality():
         c.eliminate_except = obj_eliminate_except
 
 
-
-
 _add_functionality()
 
 
-
-
 def _align_dim_type(tgt_dt, obj, tgt, obj_bigger_ok, obj_names, tgt_names):
     if None in tgt_names:
         all_nones = [None] * len(tgt_names)
@@ -630,7 +777,8 @@ def _align_dim_type(tgt_dt, obj, tgt, obj_bigger_ok, obj_names, tgt_names):
 
     tgt_idx = 0
     while tgt_idx < tgt.dim(tgt_dt):
-        tgt_name = tgt.get_dim_name(tgt_dt, tgt_idx)
+        tgt_id = tgt.get_dim_id(tgt_dt, tgt_idx)
+        tgt_name = tgt_id.name
 
         if tgt_name in names_in_both:
             if (obj.dim(tgt_dt) > tgt_idx
@@ -657,7 +805,7 @@ def _align_dim_type(tgt_dt, obj, tgt, obj_bigger_ok, obj_names, tgt_names):
                 tgt_idx += 1
         else:
             obj = obj.insert_dims(tgt_dt, tgt_idx, 1)
-            obj = obj.set_dim_name(tgt_dt, tgt_idx, tgt_name)
+            obj = obj.set_dim_id(tgt_dt, tgt_idx, tgt_id)
             tgt_idx += 1
 
     if tgt_idx < obj.dim(tgt_dt) and not obj_bigger_ok:
@@ -666,8 +814,6 @@ def _align_dim_type(tgt_dt, obj, tgt, obj_bigger_ok, obj_names, tgt_names):
     return obj
 
 
-
-
 def align_spaces(obj, tgt, obj_bigger_ok=False, across_dim_types=False):
     """
     Try to make the space in which *obj* lives the same as that of *tgt* by
@@ -677,6 +823,16 @@ def align_spaces(obj, tgt, obj_bigger_ok=False, across_dim_types=False):
         has more dimensions than *tgt*.
     """
 
+    have_any_param_domains = (
+            isinstance(obj, (Set, BasicSet))
+            and isinstance(tgt, (Set, BasicSet))
+            and (obj.is_params() or tgt.is_params()))
+    if have_any_param_domains:
+        if obj.is_params():
+            obj = type(obj).from_params(obj)
+        if tgt.is_params():
+            tgt = type(tgt).from_params(tgt)
+
     if isinstance(tgt, EXPR_CLASSES):
         dim_types = _CHECK_DIM_TYPES[:]
         dim_types.remove(dim_type.out)
@@ -684,11 +840,13 @@ def align_spaces(obj, tgt, obj_bigger_ok=False, across_dim_types=False):
         dim_types = _CHECK_DIM_TYPES
 
     if across_dim_types:
-        obj_names = [obj.get_dim_name(dt, i)
+        obj_names = [
+                obj.get_dim_name(dt, i)
                 for dt in dim_types
                 for i in xrange(obj.dim(dt))
                 ]
-        tgt_names = [tgt.get_dim_name(dt, i)
+        tgt_names = [
+                tgt.get_dim_name(dt, i)
                 for dt in dim_types
                 for i in xrange(tgt.dim(dt))
                 ]
@@ -705,8 +863,6 @@ def align_spaces(obj, tgt, obj_bigger_ok=False, across_dim_types=False):
     return obj
 
 
-
-
 def align_two(obj1, obj2, across_dim_types=False):
     """Align the spaces of two objects, potentially modifying both of them.
 
@@ -720,7 +876,4 @@ def align_two(obj1, obj2, across_dim_types=False):
     return (obj1, obj2)
 
 
-
-
-
 # vim: foldmethod=marker
